from .pipeable import pipeable
import re


@pipeable
def split(sep, s):
    ''' splits s into a list of substrings based on sep. analogous to s.split(sep)
    
    args:
        sep: a string used to determine the splitting locations.
        s: the string being splits

    return:
        a list of strings, resulting from splitting s at each instance of sep.
        note that sep is removed in the process.
    '''
    return s.split(sep)


def split_re(pat):
    ''' creates a pipeable function that will split a string using a regular expression based on pat.
    
    args:
        pat: a regular expression pattern used to split strings.

    return:
        a list of strings, resulting from splitting s at each instance of sep.
        note that sep is removed in the process.

    notes:
        Call: 
            To provide efficient computation, split_re returns a pipeable function.
            To perform a regular call: split_re(pat)(s) 
            Use regular syntax for piping: or s >> split_re(pat)

        Partial application: 
            Is curried and can be combined with curried functions (i.e. map)
            Example: list_of_str >> composable.strict.map(split_re("(,|\t)"))
    '''
    r = re.compile(pat)
    return pipeable(lambda s: r.split(s))




from composable.pipeable import pipeable
from itertools import islice
from functools import reduce as _original_reduce

@pipeable
def to_list(iter):
    ''' converts any iterable sequence into a list.

    Args:
        iter: an finite iterable sequence of items

    Returns:
        A list of the sequence iter, which maintains the original order
    '''
    return list(iter)


toList = to_list


@pipeable
def to_tuple(iter):
    ''' converts any iterable sequence into a tuple.

    Args:
        iter: an finite iterable sequence of items

    Returns:
        A tuple of the sequence iter, which maintains the original order
    '''
    return tuple(iter)


toTuple = to_tuple


built_in_slice = slice


@pipeable
def slice(start, stop, seq, step = 1):
    ''' Slices a seq, equivalent to seq[from_:up_to]

    Args:
        start: starting index for the slice.
        stop: final index
        step: the step size (can only be negative if seq supports slicing)
        seq: a sequence of value that supports slicing, e.g. list, tuple, or string

    Returns:
        A seq that is a subsequence of seq that starts at index `from_` 
        and includes all values up to, but not includeing `up_to`
        The returned value will be the same type as the input if slicing is supported, 
        otherwise this returns an iterable sequence.
    '''
    try:
        return seq[built_in_slice(start, stop, step)]
    except:
        if step <= 0:
            raise ValueError("Step must be positive for iterators that don't support slicing.")
        return islice(seq, start, stop, step)


@pipeable
def join(sep, L):
    ''' Joins a sequence of str by combining successive elements with sep.

    Args:
        : starting index for the slice.
        up_to: final index

    Returns:
        A list that is a subsequence of L that starts at index `from_` 
        and includes all values up to, but not includeing `up_to`
    '''
    return sep.join(L)


@pipeable
def head(n, seq):
    ''' Returns a list with the first n entries of seq, equivalent to seq[0:up_to]

    Args:
        from_: starting index for the slice.
        up_to: final index
        seq: a sequence of value that supports slicing, e.g. list, tuple, or string

    Returns:
        A seq that is a subsequence of seq that starts at index `from_` 
        and includes all values up to, but not includeing `up_to`
    '''
    return slice(0, n, 1, seq)


@pipeable
def tail(n, seq):
    ''' Returns a list with the last n entries of seq, equivalent to seq[0:up_to]

    Args:
        from_: starting index for the slice.
        up_to: final index
        seq: a sequence of value that supports slicing, e.g. list, tuple, or string

    Returns:
        A seq that is a subsequence of seq that starts at index `from_` 
        and includes all values up to, but not includeing `up_to`
    '''
    try:
        _ = len(seq)
    except:
        raise ValueError("tail needs a sequence that supports len and slicing")
    return seq[built_in_slice(n, len(seq), 1)]


@pipeable
def reduce(func, seq, init = None):
    '''reduce(function, sequence[, initial]) -> value

    Apply a function of two arguments cumulatively to the items of a sequence,
    from left to right, so as to reduce the sequence to a single value.
    For example, reduce(lambda x, y: x+y, [1, 2, 3, 4, 5]) calculates
    ((((1+2)+3)+4)+5).  If initial is present, it is placed before the items
    of the sequence in the calculation, and serves as a default when the
    sequence is empty.

    Args:
        - func: a 2 argument function with arguments representing the accumulator and next enter, respectively.
        - seq: a sequence of values
        - init: An optional initial value.  If init=None, then the first element of seq is used.
        
    Returns:
        - The result of the aggregation.
    '''
    if init is None:
        return _original_reduce(func, seq) # Uses first value as init
    else:
        return _original_reduce(func, seq, init)


from inspect import signature
from functools import reduce
from toolz import first, drop
from toolz import curry
from toolz.functoolz import Compose


class pipeable(curry):


    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


    def __rrshift__(self, other):
        """ We use the rightshift (i.e. `x >> f`) to represent piping.

        Note that this form or piping assumes a unary function call.
        Use a curried function to allow piping n-ary functions."""
        return self.__call__(other)


    def __rshift__(self, other):
        """ We use the rightshift (i.e. `x >> f`) to represent piping.

        Note that this form or piping assumes a unary function call.
        Use a curried function to allow piping n-ary functions."""
        assert callable(other), "All subsequent elements of a pipe must be callable"
        return other.__call__(self)


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['composable']

package_data = \
{'': ['*']}

install_requires = \
['python-forge>=18.6,<19.0', 'toolz>=0.11.1,<0.12.0']

setup_kwargs = {
    'name': 'composable',
    'version': '0.3.0',
    'description': 'Callable functions that be composed/piped using >> and <<',
    'long_description': 'None',
    'author': 'Todd Iverson',
    'author_email': 'tiverson@winona.edu',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6',
}


setup(**setup_kwargs)



from .pipeable import pipeable


@pipeable
def map(f, L):
    ''' applies f to all elements of L, immediately returning a list of the resulting values.

    Args:
        f: A unary function
        L: A list of values

    Returns:
        A list of the results of f applied to each element of L.
        Note that this is a strict version of the built-in map (which returns a generator).
    '''
    return [f(x) for x in L]


@pipeable
def filter(p, L):
    ''' Filters the elements of L, returning a list of all elements that pass the predicate p.

    Args:
        p: a unary boolean function used to filter the list
        L: A list of values

    Returns:
        A list of the elements of L that satisfy p.
        Note that this is a strict version of the built-in filter (which returns a generator).
    '''
    return [x for x in L if p(x)]


@pipeable
def zipWith(iter2, iter1):
    ''' zips the elements of iter1 and iter2 into a list of tuples of the form (i1, i2)

    args:
        iter2: a list of values
        iter1: a list of values

    returns:
        a list of tuples of the form (i1, i2).  this is similar to zip(iter1, iter2).
        note that
        1. the order is reversed for more natural piping, 
           that is iter1 >> zipwith(iter2) has the values of iter1 first, i.e. (i1, i2)
        2. performs strict application, where the built-in zip returns a generator.
    '''
    return [ (i1, i2) for i2, i1 in zip(iter2, iter1)]


@pipeable
def zipOnto(iter1, iter2):
    ''' zips the elements of iter1 and iter2 into a list of tuples of the form (i2, i1)

    args:
        iter1: a list of values
        iter2: a list of values

    returns:
        a list of tuples of the form (i2, i1).  this is similar to zip(iter1, iter2).
        note that
        1. the order is the same as the built-in zip, meaning piping will reverse the order
           that is iter1 >> zipOnto(iter2) has the values of iter1 first, i.e. (i2, i1)
        2. performs strict application, where the built-in zip returns a generator.
    '''
    return list(zip(iter1, iter2))

enum = enumerate

@pipeable
def enumerate(iter):
    ''' generates a list of tuples, pairing each value with its index, (index, value)

    Args:
        iter: A sequence of values

    Returns:
        A list of tuples of the form (index, value), which correspond to the index and value
        of each entry of iter.
        Note that this is a strict version of the built-in enumerate (which returns a generator).
    '''
    return list(enum(iter))
